@@ -1,37 +1,63 @@
-(**************************************************************************)
-(*                                                                        *)
-(*                         OCaml Migrate Parsetree                        *)
-(*                                                                        *)
-(*                             Frédéric Bour                              *)
-(*                         Alain Frisch, LexiFi                           *)
-(*                                                                        *)
-(*   Copyright 2017 Institut National de Recherche en Informatique et     *)
-(*     en Automatique (INRIA).                                            *)
-(*                                                                        *)
-(*   All rights reserved.  This file is distributed under the terms of    *)
-(*   the GNU Lesser General Public License version 2.1, with the          *)
-(*   special exception on linking described in the file LICENSE.          *)
-(*                                                                        *)
-(**************************************************************************)
+module From = Ast_408
+module To = Ast_407
 
 module Def = Migrate_parsetree_def
-module From = Ast_406
-module To = Ast_405
 
 let migration_error location feature =
   raise (Def.Migration_error (feature, location))
 
-let rec copy_expression :
+let rec copy_toplevel_phrase :
+  From.Parsetree.toplevel_phrase -> To.Parsetree.toplevel_phrase =
+  function
+  | From.Parsetree.Ptop_def x0 ->
+      To.Parsetree.Ptop_def (copy_structure x0)
+  | From.Parsetree.Ptop_dir
+      { From.Parsetree.pdir_name;
+        From.Parsetree.pdir_arg;
+        From.Parsetree.pdir_loc = _; } ->
+      To.Parsetree.Ptop_dir
+        (pdir_name.Location.txt,
+         (match pdir_arg with
+          | None -> To.Parsetree.Pdir_none
+          | Some arg -> copy_directive_argument arg))
+
+and copy_directive_argument :
+  From.Parsetree.directive_argument ->
+    To.Parsetree.directive_argument
+  =
+  fun
+    { From.Parsetree.pdira_desc = pdira_desc;
+      From.Parsetree.pdira_loc = _pdira_loc }
+     ->
+       (copy_directive_argument_desc pdira_desc)
+
+and copy_directive_argument_desc :
+  From.Parsetree.directive_argument_desc ->
+    To.Parsetree.directive_argument
+  =
+  function
+  | From.Parsetree.Pdir_string x0 -> To.Parsetree.Pdir_string x0
+  | From.Parsetree.Pdir_int (x0,x1) ->
+      To.Parsetree.Pdir_int (x0, (copy_option (fun x  -> x) x1))
+  | From.Parsetree.Pdir_ident x0 ->
+      To.Parsetree.Pdir_ident (copy_longident x0)
+  | From.Parsetree.Pdir_bool x0 ->
+      To.Parsetree.Pdir_bool (copy_bool x0)
+
+and copy_expression :
   From.Parsetree.expression -> To.Parsetree.expression =
   fun
     { From.Parsetree.pexp_desc = pexp_desc;
       From.Parsetree.pexp_loc = pexp_loc;
+      From.Parsetree.pexp_loc_stack = _;
       From.Parsetree.pexp_attributes = pexp_attributes }
      ->
     {
-      To.Parsetree.pexp_desc = (copy_expression_desc pexp_desc);
+      To.Parsetree.pexp_desc =
+        (copy_expression_desc pexp_desc);
       To.Parsetree.pexp_loc = (copy_location pexp_loc);
-      To.Parsetree.pexp_attributes = (copy_attributes pexp_attributes)
+      To.Parsetree.pexp_attributes =
+        (copy_attributes pexp_attributes)
     }
 
 and copy_expression_desc :
@@ -134,20 +160,21 @@ and copy_expression_desc :
           (copy_core_type x2))
   | From.Parsetree.Pexp_send (x0,x1) ->
       To.Parsetree.Pexp_send
-        ((copy_expression x0), (copy_loc (fun x  -> x) x1))
+        ((copy_expression x0),
+          (copy_loc copy_label x1))
   | From.Parsetree.Pexp_new x0 ->
       To.Parsetree.Pexp_new
         (copy_loc copy_longident x0)
   | From.Parsetree.Pexp_setinstvar (x0,x1) ->
       To.Parsetree.Pexp_setinstvar
-        ((copy_loc (fun x  -> x) x0),
+        ((copy_loc copy_label x0),
           (copy_expression x1))
   | From.Parsetree.Pexp_override x0 ->
       To.Parsetree.Pexp_override
         (List.map
            (fun x  ->
               let (x0,x1) = x  in
-              ((copy_loc (fun x  -> x) x0),
+              ((copy_loc copy_label x0),
                 (copy_expression x1))) x0)
   | From.Parsetree.Pexp_letmodule (x0,x1,x2) ->
       To.Parsetree.Pexp_letmodule
@@ -171,14 +198,23 @@ and copy_expression_desc :
         (copy_class_structure x0)
   | From.Parsetree.Pexp_newtype (x0,x1) ->
       To.Parsetree.Pexp_newtype
-        ((copy_loc (fun x  -> x) x0), (copy_expression x1))
+        ((copy_loc (fun x  -> x) x0),
+          (copy_expression x1))
   | From.Parsetree.Pexp_pack x0 ->
       To.Parsetree.Pexp_pack (copy_module_expr x0)
-  | From.Parsetree.Pexp_open (x0,x1,x2) ->
+  | From.Parsetree.Pexp_open (x0,x1) ->
+    begin match x0.From.Parsetree.popen_expr.From.Parsetree.pmod_desc with
+    | Pmod_ident lid ->
       To.Parsetree.Pexp_open
-        ((copy_override_flag x0),
-          (copy_loc copy_longident x1),
-          (copy_expression x2))
+        (copy_override_flag x0.From.Parsetree.popen_override,
+         (copy_loc copy_longident lid),
+         (copy_expression x1))
+    | Pmod_structure _ | Pmod_functor _ | Pmod_apply _
+    | Pmod_constraint _ | Pmod_unpack _ | Pmod_extension _ ->
+      migration_error x0.From.Parsetree.popen_loc Def.Pexp_open
+    end
+  | From.Parsetree.Pexp_letop { let_; ands = _; body = _; } ->
+    migration_error let_.pbop_op.loc Def.Pexp_letop
   | From.Parsetree.Pexp_extension x0 ->
       To.Parsetree.Pexp_extension (copy_extension x0)
   | From.Parsetree.Pexp_unreachable  -> To.Parsetree.Pexp_unreachable
@@ -225,6 +261,7 @@ and copy_pattern :
   fun
     { From.Parsetree.ppat_desc = ppat_desc;
       From.Parsetree.ppat_loc = ppat_loc;
+      From.Parsetree.ppat_loc_stack = _;
       From.Parsetree.ppat_attributes = ppat_attributes }
      ->
     {
@@ -303,6 +340,7 @@ and copy_core_type :
   fun
     { From.Parsetree.ptyp_desc = ptyp_desc;
       From.Parsetree.ptyp_loc = ptyp_loc;
+      From.Parsetree.ptyp_loc_stack = _;
       From.Parsetree.ptyp_attributes = ptyp_attributes }
      ->
     {
@@ -332,13 +370,7 @@ and copy_core_type_desc :
           (List.map copy_core_type x1))
   | From.Parsetree.Ptyp_object (x0,x1) ->
       To.Parsetree.Ptyp_object
-        ((List.map
-            (function
-              | From.Parsetree.Otag (x0,x1,x2) ->
-               (copy_loc (fun x  -> x) x0, (copy_attributes x1),
-                (copy_core_type x2))
-              | From.Parsetree.Oinherit _ ->
-                migration_error Location.none Def.Oinherit) x0),
+        ((List.map copy_object_field x0),
           (copy_closed_flag x1))
   | From.Parsetree.Ptyp_class (x0,x1) ->
       To.Parsetree.Ptyp_class
@@ -354,7 +386,8 @@ and copy_core_type_desc :
           (copy_option (fun x  -> List.map copy_label x) x2))
   | From.Parsetree.Ptyp_poly (x0,x1) ->
       To.Parsetree.Ptyp_poly
-        ((List.map (fun x  -> copy_loc (fun x  -> x) x) x0), (copy_core_type x1))
+        ((List.map (fun x  -> copy_loc (fun x  -> x) x) x0),
+          (copy_core_type x1))
   | From.Parsetree.Ptyp_package x0 ->
       To.Parsetree.Ptyp_package (copy_package_type x0)
   | From.Parsetree.Ptyp_extension x0 ->
@@ -373,14 +406,34 @@ and copy_package_type :
 
 and copy_row_field :
   From.Parsetree.row_field -> To.Parsetree.row_field =
-  function
-  | From.Parsetree.Rtag (x0,x1,x2,x3) ->
-      To.Parsetree.Rtag
-        ((copy_label x0.txt),
-          (copy_attributes x1), (copy_bool x2),
-          (List.map copy_core_type x3))
-  | From.Parsetree.Rinherit x0 ->
-      To.Parsetree.Rinherit (copy_core_type x0)
+  fun
+    { From.Parsetree.prf_desc = prf_desc;
+      From.Parsetree.prf_loc = _;
+      From.Parsetree.prf_attributes = prf_attributes }
+    ->
+      match prf_desc with
+      | From.Parsetree.Rtag (x0, x1, x2) ->
+        To.Parsetree.Rtag ((copy_loc copy_label x0),
+                                (copy_attributes prf_attributes),
+                                (copy_bool x1),
+                                (List.map copy_core_type x2))
+      | From.Parsetree.Rinherit x0 ->
+        To.Parsetree.Rinherit (copy_core_type x0)
+
+and copy_object_field :
+  From.Parsetree.object_field -> To.Parsetree.object_field =
+  fun
+    { From.Parsetree.pof_desc = pof_desc;
+      From.Parsetree.pof_loc = _;
+      From.Parsetree.pof_attributes = pof_attributes }
+    ->
+      match pof_desc with
+      | From.Parsetree.Otag (x0, x1) ->
+        To.Parsetree.Otag ((copy_loc copy_label x0),
+                                (copy_attributes pof_attributes),
+                                (copy_core_type x1))
+      | From.Parsetree.Oinherit x0 ->
+        To.Parsetree.Oinherit (copy_core_type x0)
 
 and copy_attributes :
   From.Parsetree.attributes -> To.Parsetree.attributes =
@@ -388,10 +441,13 @@ and copy_attributes :
 
 and copy_attribute :
   From.Parsetree.attribute -> To.Parsetree.attribute =
-  fun x  ->
-    let (x0,x1) = x  in
-    ((copy_loc (fun x  -> x) x0),
-      (copy_payload x1))
+  fun
+    { From.Parsetree.attr_name = attr_name;
+      From.Parsetree.attr_payload = attr_payload;
+      From.Parsetree.attr_loc = _ }
+    ->
+      ((copy_loc (fun x  -> x) attr_name),
+       (copy_payload attr_payload))
 
 and copy_payload :
   From.Parsetree.payload -> To.Parsetree.payload =
@@ -448,7 +504,10 @@ and copy_structure_item_desc :
         (copy_type_extension x0)
   | From.Parsetree.Pstr_exception x0 ->
       To.Parsetree.Pstr_exception
-        (copy_extension_constructor x0)
+        (let e = copy_extension_constructor
+                   x0.From.Parsetree.ptyexn_constructor in
+         { e with pext_attributes = e.pext_attributes @ (copy_attributes x0.ptyexn_attributes) }
+        )
   | From.Parsetree.Pstr_module x0 ->
       To.Parsetree.Pstr_module
         (copy_module_binding x0)
@@ -459,8 +518,17 @@ and copy_structure_item_desc :
       To.Parsetree.Pstr_modtype
         (copy_module_type_declaration x0)
   | From.Parsetree.Pstr_open x0 ->
+    begin match x0.From.Parsetree.popen_expr.From.Parsetree.pmod_desc with
+    | Pmod_ident lid ->
       To.Parsetree.Pstr_open
-        (copy_open_description x0)
+        { To.Parsetree.popen_lid = (copy_loc copy_longident lid);
+          To.Parsetree.popen_override = (copy_override_flag x0.From.Parsetree.popen_override);
+          To.Parsetree.popen_loc = (copy_location x0.From.Parsetree.popen_loc);
+          To.Parsetree.popen_attributes = (copy_attributes x0.From.Parsetree.popen_attributes); }
+    | Pmod_structure _ | Pmod_functor _ | Pmod_apply _
+    | Pmod_constraint _ | Pmod_unpack _ | Pmod_extension _ ->
+      migration_error x0.From.Parsetree.popen_loc Def.Pexp_open
+    end
   | From.Parsetree.Pstr_class x0 ->
       To.Parsetree.Pstr_class
         (List.map copy_class_declaration x0)
@@ -540,8 +608,11 @@ and copy_class_expr_desc :
           (copy_class_type x1))
   | From.Parsetree.Pcl_extension x0 ->
       To.Parsetree.Pcl_extension (copy_extension x0)
-  | From.Parsetree.Pcl_open (_, loc, _) ->
-      migration_error loc.Location.loc Def.Pcl_open
+  | From.Parsetree.Pcl_open (x0,x1) ->
+    To.Parsetree.Pcl_open
+      ((copy_override_flag x0.From.Parsetree.popen_override),
+       (copy_loc copy_longident x0.From.Parsetree.popen_expr),
+       (copy_class_expr x1))
 
 and copy_class_structure :
   From.Parsetree.class_structure -> To.Parsetree.class_structure =
@@ -578,17 +649,18 @@ and copy_class_field_desc :
       To.Parsetree.Pcf_inherit
         ((copy_override_flag x0),
           (copy_class_expr x1),
-          (copy_option (copy_loc (fun x  -> x)) x2))
+          (copy_option (fun x  -> copy_loc (fun x  -> x) x)
+             x2))
   | From.Parsetree.Pcf_val x0 ->
       To.Parsetree.Pcf_val
         (let (x0,x1,x2) = x0  in
-         ((copy_loc (fun x  -> x) x0),
+         ((copy_loc copy_label x0),
            (copy_mutable_flag x1),
            (copy_class_field_kind x2)))
   | From.Parsetree.Pcf_method x0 ->
       To.Parsetree.Pcf_method
         (let (x0,x1,x2) = x0  in
-         ((copy_loc (fun x  -> x) x0),
+         ((copy_loc copy_label x0),
            (copy_private_flag x1),
            (copy_class_field_kind x2)))
   | From.Parsetree.Pcf_constraint x0 ->
@@ -724,16 +796,14 @@ and copy_with_constraint :
       To.Parsetree.Pwith_module
         ((copy_loc copy_longident x0),
           (copy_loc copy_longident x1))
-  | From.Parsetree.Pwith_typesubst ({ txt = Longident.Lident _; _ }, x0) ->
+  | From.Parsetree.Pwith_typesubst (x0,x1) ->
       To.Parsetree.Pwith_typesubst
-        (copy_type_declaration x0)
-  | From.Parsetree.Pwith_modsubst ({ txt = Longident.Lident x0; loc },x1) ->
+        ((copy_loc copy_longident x0),
+          (copy_type_declaration x1))
+  | From.Parsetree.Pwith_modsubst (x0,x1) ->
       To.Parsetree.Pwith_modsubst
-        ({ txt = x0; loc }, (copy_loc copy_longident x1))
-  | From.Parsetree.Pwith_typesubst ({ loc; _ }, _x0) ->
-      migration_error loc Pwith_typesubst_longident
-  | From.Parsetree.Pwith_modsubst ({ loc; _ },_x1) ->
-      migration_error loc Pwith_modsubst_longident
+        ((copy_loc copy_longident x0),
+          (copy_loc copy_longident x1))
 
 and copy_signature :
   From.Parsetree.signature -> To.Parsetree.signature =
@@ -763,15 +833,25 @@ and copy_signature_item_desc :
       To.Parsetree.Psig_type
         ((copy_rec_flag x0),
           (List.map copy_type_declaration x1))
+  | From.Parsetree.Psig_typesubst x0 ->
+    let x0_loc =
+      match x0 with
+      | [] -> Location.none
+      | { From.Parsetree.ptype_loc; _ } :: _ -> ptype_loc in
+    migration_error x0_loc Def.Psig_typesubst
   | From.Parsetree.Psig_typext x0 ->
       To.Parsetree.Psig_typext
         (copy_type_extension x0)
   | From.Parsetree.Psig_exception x0 ->
       To.Parsetree.Psig_exception
-        (copy_extension_constructor x0)
+        (let e = copy_extension_constructor
+                   x0.From.Parsetree.ptyexn_constructor in
+         {e with pext_attributes = e.pext_attributes @ (copy_attributes x0.ptyexn_attributes) })
   | From.Parsetree.Psig_module x0 ->
       To.Parsetree.Psig_module
         (copy_module_declaration x0)
+  | From.Parsetree.Psig_modsubst x0 ->
+      migration_error x0.pms_loc Def.Psig_modsubst
   | From.Parsetree.Psig_recmodule x0 ->
       To.Parsetree.Psig_recmodule
         (List.map copy_module_declaration x0)
@@ -842,8 +922,11 @@ and copy_class_type_desc :
           (copy_class_type x2))
   | From.Parsetree.Pcty_extension x0 ->
       To.Parsetree.Pcty_extension (copy_extension x0)
-  | From.Parsetree.Pcty_open (_, loc, _) ->
-      migration_error loc.Location.loc Def.Pcty_open
+  | From.Parsetree.Pcty_open (x0,x1) ->
+    To.Parsetree.Pcty_open
+      ((copy_override_flag x0.From.Parsetree.popen_override),
+       (copy_loc copy_longident x0.From.Parsetree.popen_expr),
+       (copy_class_type x1))
 
 and copy_class_signature :
   From.Parsetree.class_signature -> To.Parsetree.class_signature =
@@ -883,13 +966,15 @@ and copy_class_type_field_desc :
   | From.Parsetree.Pctf_val x0 ->
       To.Parsetree.Pctf_val
         (let (x0,x1,x2,x3) = x0  in
-         (copy_loc (fun x  -> x) x0, (copy_mutable_flag x1),
+         ((copy_loc copy_label x0),
+           (copy_mutable_flag x1),
            (copy_virtual_flag x2),
            (copy_core_type x3)))
   | From.Parsetree.Pctf_method x0 ->
       To.Parsetree.Pctf_method
         (let (x0,x1,x2,x3) = x0  in
-         (copy_loc (fun x  -> x) x0, (copy_private_flag x1),
+         ((copy_loc copy_label x0),
+           (copy_private_flag x1),
            (copy_virtual_flag x2),
            (copy_core_type x3)))
   | From.Parsetree.Pctf_constraint x0 ->
@@ -906,6 +991,14 @@ and copy_extension :
   From.Parsetree.extension -> To.Parsetree.extension =
   fun x  ->
     let (x0,x1) = x  in
+    let x1 =
+      match x0.txt with
+        | "ocaml.error" | "error" ->
+          begin match x1 with
+          | PStr (hd :: tl) -> From.Parsetree.PStr (hd :: hd :: tl)
+          | _ -> x1
+          end
+        | _ -> x1 in
     ((copy_loc (fun x  -> x) x0),
       (copy_payload x1))
 
@@ -975,20 +1068,15 @@ and copy_include_infos :
 and copy_open_description :
   From.Parsetree.open_description -> To.Parsetree.open_description =
   fun
-    { From.Parsetree.popen_lid = popen_lid;
+    { From.Parsetree.popen_expr = popen_expr;
       From.Parsetree.popen_override = popen_override;
       From.Parsetree.popen_loc = popen_loc;
       From.Parsetree.popen_attributes = popen_attributes }
-     ->
-    {
-      To.Parsetree.popen_lid =
-        (copy_loc copy_longident popen_lid);
-      To.Parsetree.popen_override =
-        (copy_override_flag popen_override);
-      To.Parsetree.popen_loc = (copy_location popen_loc);
-      To.Parsetree.popen_attributes =
-        (copy_attributes popen_attributes)
-    }
+    ->
+      { To.Parsetree.popen_lid = (copy_loc copy_longident popen_expr);
+        To.Parsetree.popen_override = (copy_override_flag popen_override);
+        To.Parsetree.popen_loc = (copy_location popen_loc);
+        To.Parsetree.popen_attributes = (copy_attributes popen_attributes); }
 
 and copy_override_flag :
   From.Asttypes.override_flag -> To.Asttypes.override_flag =
@@ -1036,6 +1124,21 @@ and copy_module_declaration :
       To.Parsetree.pmd_loc = (copy_location pmd_loc)
     }
 
+(* and copy_type_exception :
+  From.Parsetree.type_exception -> To.Parsetree.type_exception =
+  fun
+    { From.Parsetree.ptyexn_constructor = ptyexn_constructor;
+      From.Parsetree.ptyexn_loc = ptyexn_loc;
+      From.Parsetree.ptyexn_attributes = ptyexn_attributes }
+     ->
+    {
+      To.Parsetree.ptyexn_constructor =
+        (copy_extension_constructor ptyexn_constructor);
+      To.Parsetree.ptyexn_loc = (copy_location ptyexn_loc);
+      To.Parsetree.ptyexn_attributes =
+        (copy_attributes ptyexn_attributes)
+    }*)
+
 and copy_type_extension :
   From.Parsetree.type_extension -> To.Parsetree.type_extension =
   fun
@@ -1043,6 +1146,7 @@ and copy_type_extension :
       From.Parsetree.ptyext_params = ptyext_params;
       From.Parsetree.ptyext_constructors = ptyext_constructors;
       From.Parsetree.ptyext_private = ptyext_private;
+      From.Parsetree.ptyext_loc = _;
       From.Parsetree.ptyext_attributes = ptyext_attributes }
      ->
     {
@@ -1297,27 +1401,6 @@ and copy_location : Location.t -> Location.t = fun x -> x
 
 and copy_bool : bool -> bool = function | false  -> false | true  -> true
 
-let rec copy_toplevel_phrase :
-  From.Parsetree.toplevel_phrase -> To.Parsetree.toplevel_phrase =
-  function
-  | From.Parsetree.Ptop_def x0 ->
-      To.Parsetree.Ptop_def (copy_structure x0)
-  | From.Parsetree.Ptop_dir (x0,x1) ->
-      To.Parsetree.Ptop_dir
-        (x0, (copy_directive_argument x1))
-
-and copy_directive_argument :
-  From.Parsetree.directive_argument -> To.Parsetree.directive_argument =
-  function
-  | From.Parsetree.Pdir_none  -> To.Parsetree.Pdir_none
-  | From.Parsetree.Pdir_string x0 -> To.Parsetree.Pdir_string x0
-  | From.Parsetree.Pdir_int (x0,x1) ->
-      To.Parsetree.Pdir_int (x0, (copy_option (fun x  -> x) x1))
-  | From.Parsetree.Pdir_ident x0 ->
-      To.Parsetree.Pdir_ident (copy_longident x0)
-  | From.Parsetree.Pdir_bool x0 ->
-      To.Parsetree.Pdir_bool (copy_bool x0)
-
 let copy_cases x = List.map copy_case x
 let copy_pat = copy_pattern
 let copy_expr = copy_expression
